from ..middlewares.rest_middlewares import (
    HeaderValidationMiddleware,
    ExceptionMiddleware,
    ContextSetter,
)

from ..middlewares.message_middlewares import MessageMiddleware

from ..middlewares.rate_limiter_middleware import RateLimiterGuard



from typing import Any


class SingletonMeta(type):
    _instances = {}

    def __call__(cls, *args: Any, **kwargs: Any):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class Para_SingletonMeta(type):
    _instances = {}

    def __call__(cls, *args: Any, **kwargs: Any):
        key = (cls, args, frozenset(kwargs.items()))
        if key not in cls._instances:
            cls._instances[key] = super().__call__(*args, **kwargs)
        return cls._instances[key]



import re
from fastapi.responses import StreamingResponse
from abc import ABC, abstractmethod
import json
from typing import Generic, Mapping, Optional, Dict, List, Any, TypeVar


from ...clients.bedrock_client.local_interfaces_typehinter import (
    LLM_ClientPayload_TH,
)

from .bedrock_client import GenerateBedrockResponse
from ...interfaces.interfaces_th import (
    LLM_HyperParameters_TH,
    LLM_PromptTemplates_TH,
)
from .foundation_models import FoundationModels
from ...logging.base_logger import APP_LOGGER


T = TypeVar("T", bound=Mapping)


class LLMOperation(ABC, Generic[T]):

    def get_json_response(self, string: str):
        json_match = re.search(r"{.*}", string, re.DOTALL)
        if json_match:
            json_string = json_match.group(0)
            try:
                json_object = json.loads(json_string)
                return json_object
            except json.JSONDecodeError as e:
                APP_LOGGER.error(f"Error decoding JSON: {e}")
                return None
        APP_LOGGER.warning("No JSON object found in the string.")
        return None

    def __add_roles_to_prompts(
        self,
        user_prompt: str,
        system_prompt: Optional[str] = None,
    ) -> List[Dict[str, str]]:
        """Constructs a list of message dictionaries for LLM input."""
        messages = (
            [{"role": "system", "content": system_prompt}] if system_prompt else []
        )
        messages.append(dict(role="user", content=user_prompt))
        return messages

    def construct_llmclient_payload(
        self,
        system_prompt: Optional[str],
        user_prompt: str,
        model_hyperparameters: LLM_HyperParameters_TH,
        modelId: FoundationModels,
    ) -> LLM_ClientPayload_TH:
        messages = self.__add_roles_to_prompts(
            user_prompt=user_prompt, system_prompt=system_prompt
        )
        return LLM_ClientPayload_TH(
            messages=messages,
            model_hyperparameters=model_hyperparameters,
            modelId=modelId,
        )

    def call_llmclient(
        self, payload: LLM_ClientPayload_TH, streaming: bool = False
    ) -> str:

        llm_response = GenerateBedrockResponse(
            payload=payload, streaming=streaming
        ).execute()
        if llm_response is not None and not isinstance(llm_response, StreamingResponse):
            summary: str = str(llm_response.get("llm_response")).strip()
            return summary
        else:
            raise ValueError(
                "Expected a non-streaming response and received either None or a StreamingResponse"
            )

    def call_llmclient_with_retries(
        self,
        payload: LLM_ClientPayload_TH,
        max_retries: int = 3,
    ):
        for attempt in range(max_retries):
            llm_response = self.call_llmclient(payload=payload).strip()
            APP_LOGGER.info(llm_response)
            json_response = self.get_json_response(llm_response)
            APP_LOGGER.info(json_response)
            if json_response is not None:
                return json_response
            APP_LOGGER.warning(f"Attempt {attempt + 1}: Failed to extract JSON.")
        raise ValueError(f"Failed to extract JSON after {max_retries} attempts.")

    @abstractmethod
    def execute_llm_operation(self, payload: T, modelId: FoundationModels) -> Any:
        raise NotImplementedError("Subclasses must implement execute_llm_operation.")


class BaseAIEventsConfig(ABC):
    """Abstract base class for all AI event configurations subtypes."""

    @staticmethod
    @abstractmethod
    def MODEL_HYPERPARAMETERS(
        *args, **kwargs
    ) -> Dict[FoundationModels, LLM_HyperParameters_TH]:
        """Abstract property for model parameters. Must be implemented by subclasses."""
        raise NotImplementedError("Subclasses must implement MODEL_HYPERPARAMETERS")

    @staticmethod
    @abstractmethod
    def MODEL_PROMPTS(
        *args, **kwargs
    ) -> Dict[FoundationModels, LLM_PromptTemplates_TH]:
        """Abstract property for prompts. Must be implemented by subclasses."""
        raise NotImplementedError("Subclasses must implement MODEL_PROMPTS")

    @staticmethod
    @abstractmethod
    def BASE_OUTPUT_TOKENS() -> int:
        """Abstract property for output tokens. Must be implemented by subclasses."""
        raise NotImplementedError("Subclasses must implement BASE_OUTPUT_TOKENS")



from typing import TypedDict, List, Dict

from ...clients.bedrock_client.foundation_models import (
    FoundationModels,
)

from ...interfaces.interfaces_th import LLM_HyperParameters_TH


class LLM_ClientPayload_TH(TypedDict):
    messages: List[Dict[str, str]]
    model_hyperparameters: LLM_HyperParameters_TH
    modelId: FoundationModels



import boto3
import json
from typing import List, cast
from decouple import config
from ...ioc.singleton import SingletonMeta


class EmbeddingGenerator(metaclass=SingletonMeta):
    def __init__(self) -> None:
        self.__client = boto3.client(
            "bedrock-runtime",
            region_name=str(config("AWS_REGION_NAME")),
            aws_access_key_id=str(config("AWS_ACCESS_KEY")),
            aws_secret_access_key=str(config("AWS_SECRET_ACCESS_KEY")),
        )
        self.__model_id = "amazon.titan-embed-text-v2:0"
        self.__dimensions = 256
        self.__normalize = True

    def generate_embedding(self, input_text: str) -> List[float]:
        native_request = dict(
            inputText=input_text,
            dimensions=self.__dimensions,
            normalize=self.__normalize,
        )
        response = self.__client.invoke_model(
            modelId=self.__model_id, body=json.dumps(native_request)
        )
        bedrock_response = json.loads(response["body"].read())
        embedding = bedrock_response["embedding"]
        return cast(List[float], embedding)



from abc import ABC, abstractmethod
import asyncio
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
import json
from typing import Any, Dict, Generic, Type, TypeVar
from fastapi import FastAPI
from pydantic import BaseModel

from ..interfaces.interfaces_th import (
    Headers_TH,
    SQSClientCallBackResponse_TH,
)
from ..middlewares.message_middlewares import (
    MessageMiddleware,
    ExceptionLogger,
)


T = TypeVar("T", bound=BaseModel)


class AsyncTaskHandler(ABC, Generic[T]):
    pydantic_model_class: Type[T]

    def __init__(self, app: FastAPI) -> None:
        self.app = app

    @classmethod
    @abstractmethod
    def execute_business_logic(cls, payload: T) -> None:
        """Execute the specific business logic for the action."""
        raise NotImplementedError("Subclasses must implement execute_business_logic")

    @classmethod
    def context_setter_and_execute_payload(
        cls, payload: Dict[str, Any], headers: Headers_TH
    ) -> None:
        """Set the context and execute the business logic."""
        MessageMiddleware.validate_set_context(
            payload=payload,
            headers=headers,
            payload_pydantic_model=cls.pydantic_model_class,
        )
        cls.execute_business_logic(payload=cls.pydantic_model_class(**payload))

    async def handle_with_process_pool(
        self, payload: Dict[str, Any], headers: Headers_TH
    ) -> str:
        process_pool: ProcessPoolExecutor = self.app.state.process_pool
        correlation_id = headers["correlationid"]
        try:
            future = process_pool.submit(
                self.context_setter_and_execute_payload,
                payload=payload,
                headers=headers,
            )
            await asyncio.wrap_future(future)
            response = SQSClientCallBackResponse_TH(
                allSuccess=True, correlationid=correlation_id
            )

        except Exception:
            ExceptionLogger.log_exception(self, correlation_id=correlation_id)
            response = SQSClientCallBackResponse_TH(
                allSuccess=False, correlationid=correlation_id
            )
        return json.dumps(response)

    async def handle_with_thread_pool(
        self, payload: Dict[str, Any], headers: Headers_TH
    ) -> str:
        thread_pool: ThreadPoolExecutor = self.app.state.thread_pool
        correlation_id = headers["correlationid"]
        try:
            future = thread_pool.submit(
                self.context_setter_and_execute_payload,
                payload=payload,
                headers=headers,
            )
            await asyncio.wrap_future(future)
            response = SQSClientCallBackResponse_TH(
                allSuccess=True, correlationid=correlation_id
            )

        except Exception:
            ExceptionLogger.log_exception(self, correlation_id=correlation_id)
            response = SQSClientCallBackResponse_TH(
                allSuccess=False, correlationid=correlation_id
            )
        return json.dumps(response)



from fastapi import HTTPException
from typing import Optional, cast
from redis.exceptions import RedisError

from ..clients.redis_client.async_redisclient import (
    AsyncRedisClient,
)

from ..clients.redis_client.enums import RedisExpiryEnums


class RateLimiterGuard:
    def __init__(self):
        self.redis_client = AsyncRedisClient()

    async def implement(
        self,
        key: str,
        cache_expiry: RedisExpiryEnums,
        max_calls: int,
        raise429: bool = True,
    ) -> Optional[bool]:
        """Checks if the request exceeds the rate limit."""
        try:
            match cache_expiry:
                case RedisExpiryEnums.ONE_MIN_EXPIRY:
                    cache_key = f"{key}:ONE_MIN_EXPIRY"
                case RedisExpiryEnums.ONE_DAY_EXPIRY:
                    cache_key = f"{key}:ONE_DAY_EXPIRY"
                case RedisExpiryEnums.ONE_HOUR_EXPIRY:
                    cache_key = f"{key}:ONE_HOUR_EXPIRY"
                case RedisExpiryEnums.ONE_MONTH_EXPIRY:
                    cache_key = f"{key}:ONE_MONTH_EXPIRY"
                case _:
                    raise ValueError("Invalid Redis Key.")

            current_count = cast(
                int, await self.redis_client.send_command("INCR", cache_key)
            )
            if current_count == 1:
                await self.redis_client.send_command(
                    "EXPIRE", cache_key, cache_expiry.value
                )
            if raise429:
                if current_count > max_calls:
                    raise HTTPException(
                        status_code=429,
                        detail=f"Rate limit exceeded. Max {max_calls} requests allowed.",
                    )
            else:
                return current_count > max_calls
        except RedisError:
            raise Exception("Rate limiter failed due to Redis error.")



from enum import StrEnum
from typing import Dict

"""
Configuration:
- When running locally, mock the queue names in a .env file.

Example:
<env>
embedding-service=yourname-embedding-service
evaluation-service=yourname-evaluation-service
</env>
"""


class QUEUE_NAMES(StrEnum):
    APP_001_AI_SERVICE = "app001-ai-service"


QUEUE_CONFIG: Dict[str, Dict[str, bool]] = {
    QUEUE_NAMES.APP_001_AI_SERVICE: dict(fifo=False),
}



import boto3
from typing import List, Callable, Awaitable
from typing import Dict

from .helpers import Helpers
from .message_handler import MessageHandler
from .polling import SQS_Manager
from ...ioc.singleton import SingletonMeta

# modules


class SQSInitiator(metaclass=SingletonMeta):

    def __init__(
        self,
        aws_access_key_id: str,
        aws_secret_access_key: str,
        queue_names: List[str],
        aws_region_name: str,
        queue_config: Dict[str, Dict[str, bool]],
        queue_to_message_handler_map: Dict[
            str, Dict[str, Callable[[], Awaitable[str]]]
        ],
    ) -> None:
        SQS_Manager.aws_access_key_id = aws_access_key_id
        SQS_Manager.aws_secret_access_key = aws_secret_access_key
        SQS_Manager.aws_region_name = aws_region_name

        MessageHandler.QUEUE_TO_MESSAGE_HANDLER_MAP = queue_to_message_handler_map

        self.__aws_account_id = boto3.client(
            "sts",
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
        ).get_caller_identity()["Account"]

        for queue_name in queue_names:
            queue_url = Helpers.construct_queue_url(
                queue_name=queue_name,
                region_name=aws_region_name,
                params=queue_config.get(queue_name, {}),
                aws_account_id=self.__aws_account_id,
            )
            SQS_Manager(queue_name=queue_name, queue_url=queue_url)
            MessageHandler(queue_name=queue_name)



from enum import StrEnum


class RedisExpiryEnums(StrEnum):
    ONE_MIN_EXPIRY = "60"
    ONE_HOUR_EXPIRY = "3600"
    ONE_DAY_EXPIRY = "86400"
    ONE_MONTH_EXPIRY = "2592000"  # Note  :  Assuming 30 days in a month



import boto3
import json
from typing import final, Dict
from decouple import config
from typing import TypedDict, Required, Any, cast

from ...utils.context_utils import ContextUtils

from .helpers import Helpers
from .queue_config import QUEUE_CONFIG
from ...logging.base_logger import APP_LOGGER


class Message_Payload_TypeHinter(TypedDict, total=False):
    action: Required[str]
    payload: Dict[str, Any]
    context: Dict[str, Any]


@final
class SyncSQSPusher:
    aws_access_key_id: str = cast(str, config("AWS_ACCESS_KEY"))
    aws_secret_access_key: str = cast(str, config("AWS_SECRET_ACCESS_KEY"))
    aws_region_name: str = cast(str, config("AWS_REGION_NAME", default=False))
    queue_config: Dict[str, Dict[str, bool]] = QUEUE_CONFIG

    def __construct_message_body(
        self, action: str, payload: Dict[str, Any]
    ) -> Message_Payload_TypeHinter:
        message_context = ContextUtils.get_header_details().model_dump(mode="json")
        message_payload = Message_Payload_TypeHinter(
            action=action, context=message_context, payload=payload
        )
        return message_payload

    def __init__(self, queue_name: str, action: str, payload: Dict[str, Any]) -> None:
        self.sqs_client = boto3.client(
            "sqs",
            aws_access_key_id=self.aws_access_key_id,
            aws_secret_access_key=self.aws_secret_access_key,
            region_name=self.aws_region_name,
        )
        self.aws_account_id = boto3.client(
            "sts",
            aws_access_key_id=self.aws_access_key_id,
            aws_secret_access_key=self.aws_secret_access_key,
        ).get_caller_identity()["Account"]
        self.queue_name = queue_name
        self.queue_url = Helpers.construct_queue_url(
            queue_name=queue_name,
            region_name=self.aws_region_name,
            params=self.queue_config[queue_name],
            aws_account_id=self.aws_account_id,
        )
        self.message_payload = self.__construct_message_body(
            action=action, payload=payload
        )

    def post_message(self) -> None:
        try:
            response = self.sqs_client.send_message(
                QueueUrl=self.queue_url, MessageBody=json.dumps(self.message_payload)
            )
            APP_LOGGER.info(
                f"Message sent successfully to queue '{self.queue_name}' with Message ID: { response['MessageId']} and  ->  message payload : {self.message_payload}"
            )
        except Exception as e:
            APP_LOGGER.error("Error sending message:" + str(e))
            raise e



from contextvars import ContextVar
from typing import Any

headers_context: Any = ContextVar("headers_context", default=None)
payload_context: Any = ContextVar("request_body", default=None)
request_context: Any = ContextVar("request_context", default=None)
request_params_context: Any = ContextVar("request_params", default=None)



import itertools
import asyncio
import botocore.exceptions
from aiobotocore.session import get_session
from typing import final, Dict, Any, List

from .message_handler import MessageHandler
from ...logging.base_logger import APP_LOGGER


@final
class AsyncQueueClient:
    def __init__(
        self,
        aws_access_key_id: str,
        aws_secret_access_key: str,
        aws_region_name: str,
        queue_url: str,
        queue_name: str,
    ) -> None:
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.aws_region_name = aws_region_name
        self.queue_url = queue_url
        self.queue_name = queue_name

    async def poll_messages(
        self, batch_size_of_messages: int, max_polling_duration: int
    ) -> None:

        session = get_session()
        async with session.create_client(
            "sqs",
            aws_access_key_id=self.aws_access_key_id,
            aws_secret_access_key=self.aws_secret_access_key,
            region_name=self.aws_region_name,
        ) as async_sqs_client:
            APP_LOGGER.info(f"POLLING TO --> {self.queue_url}")
            print(f"POLLING TO --> {self.queue_url} QUEUE")
            for _ in itertools.repeat(None):
                try:
                    task = asyncio.create_task(
                        async_sqs_client.receive_message(
                            QueueUrl=self.queue_url,
                            MaxNumberOfMessages=batch_size_of_messages,
                            WaitTimeSeconds=max_polling_duration,
                        )
                    )
                    response: Dict[str, Any] = await task
                except botocore.exceptions.ClientError as e:
                    APP_LOGGER.error(f"SQS Error while receiving messages: {e}")
                    continue

                messages: List[Dict[str, Any]] = response.get("Messages", [])

                if messages:
                    entries_to_delete = await MessageHandler(
                        queue_name=self.queue_name
                    ).handle_messages_in_batch(messages_batch_payload=messages)

                    if entries_to_delete:
                        message_entries = [
                            {"Id": item["Id"], "ReceiptHandle": item["ReceiptHandle"]}
                            for item in entries_to_delete
                        ]
                        correlationids = [
                            {"correlationid": item["correlationid"]}
                            for item in entries_to_delete
                        ]
                        try:
                            delete_batch_messages_response = asyncio.create_task(
                                async_sqs_client.delete_message_batch(
                                    QueueUrl=self.queue_url, Entries=message_entries
                                )
                            )
                            delete_batch_messages_response = (
                                await delete_batch_messages_response
                            )
                            APP_LOGGER.info(
                                f"Messages Deleted From SQS with {correlationids}: {str(delete_batch_messages_response)}"
                            )

                        except botocore.exceptions.ClientError as e:
                            APP_LOGGER.error(f"SQS Error while deleting messages: {e}")


@final
class SQS_Manager:
    aws_access_key_id: str
    aws_secret_access_key: str
    aws_region_name: str

    __instances: Dict[str, "SQS_Manager"] = {}

    def __new__(cls, queue_name: str, **kwargs):
        if queue_name not in cls.__instances:
            queue_url: str = str(kwargs.get("queue_url"))
            instance = super().__new__(cls)
            instance.__init__(queue_name=queue_name, queue_url=queue_url)
            cls.__instances[queue_name] = instance
        return cls.__instances[queue_name]

    def __init__(self, queue_name: str, **kwargs):
        if not hasattr(self, "initialized"):
            self.queue_name = queue_name
            self.queue_url: str = str(kwargs.get("queue_url"))
            self.async_client = AsyncQueueClient(
                aws_access_key_id=self.aws_access_key_id,
                aws_secret_access_key=self.aws_secret_access_key,
                aws_region_name=self.aws_region_name,
                queue_url=self.queue_url,
                queue_name=self.queue_name,
            )
            self.initialized: bool = True



from pymongo.collection import Collection
from typing import Optional, Type
from mongoengine import Document
from pymongo.client_session import ClientSession

from ..mongo_utils.transactions import mongo_session_context


class BaseRepository:
    @staticmethod
    def _get_session() -> Optional[ClientSession]:
        return mongo_session_context.get()

    @staticmethod
    def _get_collection(model_class: Type[Document]) -> Collection:
        return model_class._get_collection()



import requests
from decouple import config
from typing import Optional, Dict

"""
    A versatile HTTP client for making inter-service calls or external API requests.

    Attributes:
        headers (Dict[str, str]): Headers to be included in the HTTP request.
        path (str): The endpoint path for the request.
        body (Optional[Dict]): The request body for methods like POST, PATCH, PUT.
        path_params (Optional[Dict[str, str]]): Path parameters to be substituted in the URL.
        query_params (Optional[Dict[str, str]]): Query parameters to be appended to the URL.
        localhost_interservice_call (bool): If True, the request is made to localhost for development purposes.
        portNumber (Optional[str]): The port number to use when making localhost requests.
"""


class HttpClient:
    def __init__(
        self,
        headers: Dict[str, str],
        path: str,
        body=None,
        path_params: Optional[Dict[str, str]] = None,
        query_params: Optional[Dict[str, str]] = None,
        localhost_interservice_call: bool = False,
        portNumber: Optional[str] = None,
    ) -> None:
        self.__localhost_interservice_call = localhost_interservice_call
        self.__environment: str = str(config("ENVIRONMENT", default=""))
        self.__domain: str = str(config("DOMAIN", default=""))
        self.__portnumber: Optional[str] = portNumber
        self.__headers = headers
        self.__path = path
        self.__body = body
        self.__path_params = path_params
        self.__query_params = query_params

    def __construct_url(self) -> str:
        base_url = (
            f"https://{self.__domain}/"
            if not self.__localhost_interservice_call
            else f"http://localhost:{self.__portnumber}/"
        )
        url = base_url + self.__path
        if self.__path_params:
            url = url.format(**self.__path_params)
        return url

    def __send_request(self, method: str) -> requests.Response:
        url: str = self.__construct_url()
        try:
            response = requests.request(
                method=method,
                url=url,
                headers=self.__headers,
                params=self.__query_params,
                json=self.__body,
            )
            response.raise_for_status()  # Raise an exception for 4xx and 5xx status codes
            return response
        except requests.exceptions.RequestException as e:
            raise

    def get(self) -> requests.Response:
        return self.__send_request("GET")

    def post(self) -> requests.Response:
        return self.__send_request("POST")

    def patch(self) -> requests.Response:
        return self.__send_request("PATCH")

    def put(self) -> requests.Response:
        return self.__send_request("PUT")

    def delete(self) -> requests.Response:
        return self.__send_request("DELETE")
